

set -ex



pip check
pytest modin/pandas/test/dataframe/test_udf.py
pytest modin/pandas/test/dataframe/test_reduce.py modin/pandas/test/dataframe/test_window.py modin/pandas/test/dataframe/test_pickle.py modin/pandas/test/dataframe/test_binary.py modin/pandas/test/dataframe/test_iter.py
pytest -m "not exclude_in_sanity" modin/pandas/test/test_series.py
pytest -m "not exclude_in_sanity" modin/pandas/test/dataframe/test_map_metadata.py modin/pandas/test/test_expanding.py modin/pandas/test/test_reshape.py modin/pandas/test/test_general.py modin/pandas/test/test_concat.py
pytest modin/test/interchange/dataframe_protocol/test_general.py modin/pandas/test/test_concat.py modin/test/interchange/dataframe_protocol/pandas/test_protocol.py modin/pandas/test/test_rolling.py
exit 0
